# -*- mode:Python; tab-width: 3 -*-
"""Extract individual messages from UNIX mailboxes.

The module provides two classes: Mailbox, a file-like object
which allows you to iterate through the contents of a mailbox,
and Email, an object which holds the individual emails returned
by Mailbox. This module has been written with simplicity in mind,
and low memory consumption. Unless you do something bad, I estimate
maximum memory consumption as twice the memory required by the
largest email of the opened mailbox. But I'm guessing, maybe it is
lower, like on time your biggest email.

This is an example code which shows usage:

   import mailbox_reader.py
   mbox = mailbox_reader.Mailbox("whatever")
   email = mbox.read_email()
   while email:
      do_stuff_with_email_object(email)
      email.close()
      email = mbox.read_email()
   mbox.close()

This module comes with a setup.py file for installation, so running
"python setup.py install" will copy the module to your system's
python library path. You need the Distutils package to do so
(usually Python 1.5.2 users). More info at:

   http://www.python.org/sigs/distutils-sig/.

This module is known to work under python versions 1.5.2 and 2.2.3.
There are other modules implementing mailbox reader, usually
with greater complexity than what I needed. Go to the Vault
of Parnassus (http://www.vex.net/parnassus/, also linked from
http://www.python.org/) and search for `mailbox'.

This module is listed at:
   http://www.vex.net/parnassus/
   http://freshmeat.net/

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2 of the License,
or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
"""


__author__ = "Grzegorz Adam Hankiewicz"
__email__ = "gradha@users.sourceforge.net"
__date__ = "June 2003"
__version__ = "$Revision: 1.4 $"
__credits__ = ""


import string
import re
import rfc822


# A valid from line is "From return-path weekday month day time
# [timezone] year. This info was taken from http://www.mutt.org/.
_from_re = re.compile((r"From \s*[^\s]+\s+\w\w\w\s+\w\w\w\s+\d?\d\s+"
   r"\d?\d:\d\d(:\d\d)?(\s+[^\s]+)?\s+\d\d\d\d\s*$"))

         
class Mailbox:
   """Simple mailbox reader.

   Implements a simple file-like object. After opening, you are
   supposed to call the read_mail until it returns None or you are
   tired of it. The object will keep the mailbox file opened, so
   you better call close() once you are done with the mail reading.
   There is no way to go back in a Mailbox object, you would have
   to open a new one or keep yourself a copy of the messages in
   memory if you want to iterate back.
   """

   def __init__(self, filename):
      """Opens the given mailbox for filename reading."""
      self._line = self._file = None
      self._file = open(filename, "rt")

   def __del__(self):
      """Try our best to close the open file."""
      self.close()

   def read_mail(self):
      """Returns the next mail from the mailbox as Email object.

      If there are no more mails, None is returned.
      """
      headers = []
      body = []
      reading_body = 0

      if not self._line:
         self._line = self._file.readline()
         
      while self._line:
         if reading_body and _from_re.match(self._line):
            return Email(headers, body)
   
         if reading_body:
            body.append(self._line)
         elif not reading_body and not string.strip(self._line):
            """Matched the start of the body"""
            reading_body = 1
         else:
            headers.append(self._line)

         self._line = self._file.readline()

      if len(headers) and len(body):
         return Email(headers, body)

      return None

   def close(self):
      """Closes the opened mailbox forever."""
      if self._file:
         self._file.close()
         self._file = None


class Email:
   """Holds the headers and body of an individual mail.

   You can read the body lines of an email with the readline()
   method.  You can retrieve the headers and body from the
   Email object just by accessing the same attributes which hold
   them. There are two more attributes you might want to use: the
   size attribute counts the size of the email in bytes. The rfc822
   attribute is a Message instance of Python's rfc822 module. You
   could access it to get specific mail headers, though there's
   the get_header method to abbreviate.

   Note that once you are done with an email, deleting it or keeping
   zero references to it will cause a memory leak unless you call
   its close() method.  This is because the self.rfc822 object
   holds references to the Email, which turn to be cyclic.
   """
   
   def __init__(self, headers, body):
      """Construct an email holder, passing headers and body as lists.

      You should not try to build Email objects yourself, only get
      built ones from the Mailbox class.
      """
      self._reading_line = 0
      self.headers = headers
      self.body = body
      self.size = 1 # Line separating body from headers not included later
      self.header_lines = len(self.headers)
      for f in range(len(self.headers)): self.size = self.size + len(self.headers[f]) + 1
      for f in range(len(self.body)): self.size = self.size + len(self.body[f]) + 1

      # Create method to retrieve generic headers
      self.rfc822 = rfc822.Message(self)
      # Retrieve commond headers, some lowered case
      self.subject = string.strip(self.rfc822.get("subject", ""))
      self.ffrom = string.strip(self.rfc822.get("from", ""))
      self.to = string.strip(self.rfc822.get("to", ""))

   def readline(self):
      """Use this method to get lines from the message body.

      This method returns one line at a time from the body. When
      you reach the end of the body, None is returned. There's no
      way to come back to the beginning.
      """
      if self._reading_line < self.header_lines:
         result = self.headers[self._reading_line]
      elif self._reading_line == self.header_lines:
         result = "\n"
      else:
         try:
            result = self.body[self._reading_line - self.header_lines - 1]
         except IndexError:
            return None

      self._reading_line = self._reading_line + 1
      return result

   def close(self):
      """Call this to free a reference cycle, avoiding amemory leak."""
      if self.rfc822:
         self.rfc822 = None

   def get_header(self, header_name):
      """Simple wrapper around the rfc822 object held by the email.

      self.get_header(...) is equal to self.rfc822.get(..., None)
      """
      return self.rfc822.get(header_name, None)

